#include "TheDebugger.h"

#include <Memory.h>
#include <stdlib.h>
#include <string.h>


#if qTheDebugger

T_ExtDbgr gExtDbgr_Info = {0};	// define struct block for Jasik Dbgr

const long c_MA_Objs = 512;		// obj list collector increment size


#pragma segment MADebugger
void TheDbgr_Add_Object(ClassDesc::ObjectRef* objHdl)
{
	long		i,si;
	T_ObjList*	obj_list;
	
	if( gExtDbgr_Info.hMA_OBJ_List == NULL ) {
		// DPC - We could have been called during
		// startup for global objects. In this case
		// just set-up the buffer for the classes
		
		TheDbgr_Init_Ext_Dbgr();
		
		if (gExtDbgr_Info.hMA_OBJ_List == NULL) return;
	}
	
lp:	{
		obj_list = *gExtDbgr_Info.hMA_OBJ_List;
		if( obj_list->L_used == obj_list->max_objs ) {
			// squeeze out empty entries
			if( obj_list->n_holes > 0 ) {
				obj_list->n_holes = 0;
				si = 0;
				
				for(i = 0; i < obj_list->L_used; i++ ) {
					if( obj_list->objs[i] != 0 ) {
						obj_list->objs[si++] = obj_list->objs[i];
					}
				}
				
				obj_list->L_used = si;
			} else {
				obj_list->max_objs = obj_list->max_objs + c_MA_Objs;
				SetHandleSize((Handle)gExtDbgr_Info.hMA_OBJ_List, 12 + 4*obj_list->max_objs );
				obj_list = *gExtDbgr_Info.hMA_OBJ_List;
				
				if (MemError() != noErr) {
					DebugStr((StringPtr)"\pObject Collector could NOT get space to expand OBJ List");
					return;
				};
				
				goto lp;
			}
		
		}
		
		obj_list->objs[obj_list->L_used] = (long)objHdl;
		obj_list->L_used++;
	}
}

#pragma segment MADebugger
void TheDbgr_Delete_Object(ClassDesc::ObjectRef* objHdl)
{
	long	i;
	T_ObjList*	obj_list;

	if( gExtDbgr_Info.hMA_OBJ_List )
	{
		obj_list = *gExtDbgr_Info.hMA_OBJ_List;

		for(i = 0; i < obj_list->L_used; i++ )
			if( obj_list->objs[i] == (long)objHdl )
			{
				if( i == obj_list->L_used - 1 )
				{
					obj_list->L_used--;
				}
				else
				{
					obj_list->objs[i] = 0;  obj_list->n_holes++;
				}
				return;
			}
		DebugStr((StringPtr)"\pTheDbgr_Delete_Object - Object not found");
	}
}

#pragma segment MADebugger
static pascal Boolean TheDbgr_CallBack()
{
	long	ConsWR_Proc;

#if !GENERATING68K

	short proc_code[] = {  // adj stk and Call Dbgr
		0x201F,	//	move.l	(SP)+,D0	; rtn addr
		0x4267,	//	CLR	-(SP)		; setup Excpt stk frame
		0x2F00,	//	move.l	D0,-(SP)
		0x4267,	//	CLR	-(SP)
		0x7007,	//	MOVEQ	#7,D0 - isDbgrTask code
		0x2078, 0x009C,	//	MOVE.L	4*(32+7),A0	; trap #7 Handler addr
		0x4ED0	//	JMP	(A0)
	};

	enum {
	TD_ProcInfo = kPascalStackBased
		 | RESULT_SIZE(SIZE_CODE(sizeof(Boolean)))
		 | STACK_ROUTINE_PARAMETER(1, SIZE_CODE(sizeof(Ptr)))
		 | STACK_ROUTINE_PARAMETER(2, SIZE_CODE(sizeof(Ptr)))
};
	Boolean rtnVal;
	UniversalProcPtr upp = NewRoutineDescriptor((ProcPtr) &proc_code , TD_ProcInfo , kM68kISA );

	rtnVal = CallUniversalProc(upp,TD_ProcInfo,&gExtDbgr_Info,&ConsWR_Proc);
	DisposeRoutineDescriptor(upp);
	return( rtnVal );
#else

	return( TheDbgr_IsDbgrTask(&gExtDbgr_Info, &ConsWR_Proc) );

#endif
}

#if defined(__MWERKS__) && !defined(__SC__)
#pragma require_prototypes reset
#endif

#pragma segment MADebugger
void TheDbgr_Init_Ext_Dbgr()
{	
	// initialize remaining fields in gExtDbgr_Info:
	// note that n_classes and pOrderedClass have already
	// been set in ClassDesc::InitUClassDesc; all other
	// fields were zeroed at global initialization time
	gExtDbgr_Info.isBeingDebugged = false;
	gExtDbgr_Info.MA_Version = 4;
	
	if( TheDbgr_IsInstalled() && TheDbgr_CallBack() ) {
		gExtDbgr_Info.isBeingDebugged = true;

		// DPC - 25 Sep 1995
		// This routine can be called twice now since
		// global objects are being registered at
		// init time, before main.
		if (gExtDbgr_Info.hMA_OBJ_List == NULL) {
			// allocate space for an Object collection LIst
			gExtDbgr_Info.hMA_OBJ_List = (T_ObjList**) NewHandleClear((3*sizeof(long))+(c_MA_Objs*sizeof(void*)));
		
			if( gExtDbgr_Info.hMA_OBJ_List ) {
				(**gExtDbgr_Info.hMA_OBJ_List).max_objs = c_MA_Objs;
			}
		}
	}
}

#pragma segment MADebugger

typedef struct T_cl_Sort {
	short 		Cl_ID;
	short 		subCl_ID;
	ClassDesc*	cd;
} T_cl_Sort;

struct T_cl_Sort_Buf {
	long				count;
	struct T_cl_Sort	data[1];
};

typedef struct T_cl_Sort_Buf T_cl_Sort_Buf, *P_cl_Sort_Buf, **H_cl_Sort_Buf;

#define sizeof_cl_Sort_Buf(n)	(sizeof(T_cl_Sort_Buf) - sizeof(T_cl_Sort) + ((n) * sizeof(T_cl_Sort)))

static int compare_class_name1(const void* pa, const void* pb)
{
	ClassDesc* a = ((T_cl_Sort*) pa)->cd;
	ClassDesc* b = ((T_cl_Sort*) pb)->cd;

	return strcmp(a ? a->GetClassName() : "", b ? b->GetClassName() : "");
}

static ClassDesc*** NewClassListHandle(int n_class, T_cl_Sort* data)
{
	ClassDesc*** h = (ClassDesc***) NewHandle(sizeof(ClassDesc*) * n_class);
	
	if (h != nil)
	{
		ClassDesc** p = *h;
		
		for (int i = 0; i < n_class; i++)
		{
			p[i] = data[i].cd;
		}
	}
	
	return h;
}

static short find_children(
	T_cl_Sort*	class_list,
	short		num_classes,
	short		parent_id)
{
	int i = 0;
	int j = 0;
	
	while (i < num_classes)
	{
		if (class_list[i].Cl_ID == parent_id)
		{
			if (i > j)	// pull to front of list
			{
				T_cl_Sort temp = class_list[i];
				
				BlockMoveData(&class_list[j], &class_list[j+1], (i-j)*sizeof(T_cl_Sort));
				
				class_list[j] = temp;
			}
			
			i = j + 1;	
			i = j = i + find_children(&class_list[i], num_classes - i, class_list[j].subCl_ID);
		}
		else
		{
			i += 1;
		}
	}
	
	return j;
}

void TheDbgr_Adjust_ClassDescr_Ids(ClassDesc* pclassList)
{
	T_ExtDbgr* info = &gExtDbgr_Info;		// define struct block for Jasik Dbgr

	long numClasses = ClassDesc::GetNumClasses();
	
	if (numClasses > 0 && gExtDbgr_Info.n_classes != numClasses)
	{
		if (numClasses > 32767)
		{
			Debugger();
		
			return;
		}
	
		H_cl_Sort_Buf sortBuf = (H_cl_Sort_Buf) NewHandleClear(sizeof_cl_Sort_Buf(numClasses));
		
		if (sortBuf == nil)
		{
			return;
		}
		
		HLock((Handle) sortBuf);
		
		P_cl_Sort_Buf sort = *sortBuf;
		
		{
			int n_cl = 0;
			
			sort->count = numClasses;

			for (const ClassDesc* pCD = pclassList; pCD; pCD = pCD->GetNextClassDesc())
			{
				ClassID id = pCD->GetClassID();	
			
				sort->data[n_cl].subCl_ID	= id;		// id of this class is sub Class ID 
				sort->data[n_cl].cd			= const_cast<ClassDesc*>(pCD);
				
				const ClassDesc* pBC = pCD->GetFirstBaseClass();

				if ( pBC )
				{
					sort->data[n_cl].Cl_ID = pBC->GetClassID();
				}
				else
				{
					sort->data[n_cl].Cl_ID = 0;
				}

				n_cl += 1;
			}
			
			if (n_cl != numClasses)
			{
				Debugger();
			}
		}

		gExtDbgr_Info.n_classes = numClasses;

		// sort class table by name

		qsort(&sort->data[0], numClasses, sizeof(T_cl_Sort), compare_class_name1);

		gExtDbgr_Info.MA_ClassTblPtr = NewClassListHandle(numClasses, sort->data);
		
		// now sort it by class (depth first)
		
		find_children(sort->data, numClasses, 0);
		
		// and reassign class id's to match index in list
		
		for (int i = numClasses; i > 0; --i)
		{
			sort->data[i-1].cd->SetClassID(i);
		}
		
		gExtDbgr_Info.pOrderedClass = NewClassListHandle(numClasses, sort->data);

		DisposeHandle((Handle) sortBuf);
	}
}

#endif	// #if qTheDebugger
