#pragma once

#include "ClassDesc.h"

// The following routines come to MacApp courtesy of Steve Jasik, and provide
// support for Jasik's The Debugger from within the framework.

#if qDebug || qTheDebugger

	// The Debugger is installed if Exception Vector at 0x38 points to a PEA instruction

	inline bool TheDbgr_IsInstalled(void) { return (** (short **) 0x38) == 0x487A; }

#endif

#if qTheDebugger

EXTERN_API( void )
TheDbgr_StoreWord				(short					value,								// MOVEQ #8,D0, TRAP #7
								 void*					addr)								TWOWORDINLINE(0x7008, 0x4E47);

EXTERN_API( void )
TheDbgr_StoreLong				(long					value,								// MOVEQ #9,D0, TRAP #7
								 void*					addr)								TWOWORDINLINE(0x7009, 0x4E47);

#if PRAGMA_STRUCT_ALIGN
	#pragma options align=mac68k
#elif PRAGMA_STRUCT_PACKPUSH
	#pragma pack(push, 2)
#elif PRAGMA_STRUCT_PACK
	#pragma pack(2)
#endif

	typedef struct T_ObjList
	{
		long max_objs;
		long L_used;
		long n_holes;
		long objs[1024]; 	// ptrs to the objects, longint indexing 
	} T_ObjList;

	//	NOTE: Boolean MUST BE A single byte

	typedef struct T_ExtDbgr {
		Boolean			isBeingDebugged;	// True if a Debugger & a .SYM file present
		Byte			filler1;
		Byte			MA_Version;			// = 2 for MacApp 3.1 else 0
		Byte			filler2;
		short			n_classes;			// number of Object classes
		ClassDesc***	pOrderedClass;		// the tbl of ordered classes, Handle if MA3.1
		long			hCLASINFO;			// Handle to %_CLASINFO - MA3.0 and earlier
		T_ObjList**		hMA_OBJ_List;		// Handle to list of MacApp Objects
		long			MA_JMPtoTrap;		// @  %_JMP TO TRAP - MA3.0 and earlier 
		long			MA_FailNObj;		// @ FailNonObject - MA3.0 and earlier
		long			MA_DiscMeth;		// @ PDisciplineMethodCalls:BOOLEAN
		ClassDesc***	MA_ClassTblPtr;		// Class Tbl Ptr - MA3.0 and earlier
		long			MA_SelProcPtr;		// SelectorProcTable - MA3.0 and earlier
	} T_ExtDbgr;

	extern T_ExtDbgr gExtDbgr_Info;

#if PRAGMA_STRUCT_ALIGN
	#pragma options align=reset
#elif PRAGMA_STRUCT_PACKPUSH
	#pragma pack(pop)
#elif PRAGMA_STRUCT_PACK
	#pragma pack()
#endif

#if !qPowerPC && !defined(powerc)
	#if defined(__MWERKS__) && defined(__PowerPlant__)

		// This requires some low-level debugger to be installed
		// either macsbug or jasik. W/o either bad things happen,
		// but you would figure that wouldn't you.
		inline asm Boolean TheDbgr_IsDbgrTask(T_ExtDbgr* comBlk, long* ConsWR_Proc) : __D0
		{
				moveq	#7, d0
				trap	#7
			
				cmpi.l	#7, d0		// Is the debugger installed?
				bne.s	_dbgr
				moveq	#0, d0		// Failure condition w/o debugger
				rts					// Try to return to the right function
			
		_dbgr:	adda.l	#4, a7		// Futz with the stack ptr
				adda.l	#4, a7
				movem.l	(a7)+, a6	// Adjust the frame ptr
				rts
		};

	#else

		//	Debugger Callback - old was VAR ConsWR_Proc:Longint
		inline pascal Boolean TheDbgr_IsDbgrTask(T_ExtDbgr* comBlk, long* ConsWR_Proc) = 
		{	
			0x7007,		// MOVEQ #7,D0
			0x4E47		// TRAP #7
		};

	#endif /* defined(__MWERKS__) && defined(__PowerPlant__) */
#endif /* !qPowerPC && !defined(powerc) */

//	procedures for the object list collector

void TheDbgr_Add_Object(ClassDesc::ObjectRef* objHdl);
void TheDbgr_Delete_Object(ClassDesc::ObjectRef* objHdl);
void TheDbgr_Init_Ext_Dbgr(void);
void TheDbgr_Adjust_ClassDescr_Ids(ClassDesc* pclassList);

#endif	// #if qTheDebugger
